/**********************************************************************
Copyright 1999 by Object Craft

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Object Craft not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

OBJECT CRAFT DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL OBJECT CRAFT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
**********************************************************************/

#include "mssqldb.h"

#define maxv(a,b) ((a) > (b) ? (a) : (b))

PyTypeObject NumericType;

static struct PyMethodDef Numeric_methods[] = {
    { NULL }			/* sentinel */
};

int numeric_as_string(PyObject *obj, char *text)
{
    int len;

    len = dbconvert(NULL,
		    SQLNUMERIC, (BYTE*)&((NumericObj*)obj)->num, -1,
		    SQLCHAR, text, NUMERIC_LEN);
    if (len > 0)
	text[len] = '\0';
    return len;
}

PyObject *numeric_alloc(DBNUMERIC *num)
{
    NumericObj *self;

    self = PyObject_NEW(NumericObj, &NumericType);
    if (self == NULL)
	return NULL;

    memcpy(&self->num, num, sizeof(self->num));
    return (PyObject*)self;
}

#ifdef _WIN32
PyObject *numeric_from_int(long num, int precision, int scale)
{
    DBNUMERIC numeric_value;

    if (dbconvert(NULL,
		  SQLINT4, (BYTE*)&num, -1,
		  SQLNUMERIC, (BYTE*)&numeric_value, -1) <= 0) {
	PyErr_SetString(PyExc_TypeError, "numeric from int conversion failed");
	return NULL;
    }
    return numeric_alloc(&numeric_value);
}
#else
PyObject *numeric_from_int(long num, int precision, int scale)
{
    DBNUMERIC numeric_value;
    DBTYPEINFO info;

    if (precision < 0)
	info.precision = 16;
    else
	info.precision = precision;
    if (scale < 0)
	info.scale = 0;
    else
	info.scale = scale;
    if (dbconvert_ps(NULL,
		     SQLINT4, (BYTE*)&num, -1,
		     SQLNUMERIC, (BYTE*)&numeric_value, -1,
		     &info) <= 0) {
	PyErr_SetString(PyExc_TypeError, "numeric from int conversion failed");
	return NULL;
    }
    return numeric_alloc(&numeric_value);
}
#endif

PyObject *Numeric_FromInt(PyObject *obj, int precision, int scale)
{
    return numeric_from_int(PyInt_AsLong(obj), precision, scale);
}

#ifdef _WIN32
PyObject *Numeric_FromString(PyObject *obj, int precision, int scale)
{
    DBNUMERIC numeric_value;
    char *str = PyString_AsString(obj);
    
    if (dbconvert(NULL,
		  SQLCHAR, str, -1,
		  SQLNUMERIC, (BYTE*)&numeric_value, -1) <= 0) {
	PyErr_SetString(PyExc_TypeError, "numeric from string conversion failed");
	return NULL;
    }
    return numeric_alloc(&numeric_value);
}
#else
PyObject *Numeric_FromString(PyObject *obj, int precision, int scale)
{
    DBNUMERIC numeric_value;
    DBTYPEINFO info;
    char *str = PyString_AsString(obj);
    char *dp = strchr(str, '.');
    int len = strlen(str);
    
    if (precision < 0) {
	precision = len;
	if (precision > DB_MAX_PREC)
	    precision = DB_MAX_PREC;
    }
    info.precision = precision;
    if (scale < 0) {
	if (dp) {
	    int decimals = len - (dp - str) - 1;
	    if (decimals > DB_MAX_SCALE)
		decimals = DB_MAX_SCALE;
	    scale = decimals;
	} else
	    scale = 0;
    }
    info.scale = scale;
    if (dbconvert_ps(NULL,
		     SQLCHAR, str, -1,
		     SQLNUMERIC, (BYTE*)&numeric_value, -1,
		     &info) <= 0) {
	PyErr_SetString(PyExc_TypeError, "numeric from string conversion failed");
	return NULL;
    }
    return numeric_alloc(&numeric_value);
}
#endif

#ifdef _WIN32
PyObject *Numeric_FromLong(PyObject *obj, int precision, int scale)
{
    DBNUMERIC numeric_value;
    DBINT conv_result;
    PyObject *strobj = PyObject_Str(obj);
    char *str;
    int num_digits;

    if (strobj == NULL)
	return NULL;
    str = PyString_AsString(strobj);
    num_digits = strlen(str);
    if (str[num_digits - 1] == 'L')
	num_digits--;
    conv_result = dbconvert(NULL,
			    SQLCHAR, str, num_digits,
			    SQLNUMERIC, (BYTE*)&numeric_value, -1);
    Py_DECREF(strobj);
    if (conv_result <= 0) {
	PyErr_SetString(PyExc_TypeError, "numeric from long conversion failed");
	return NULL;
    }
    return numeric_alloc(&numeric_value);
}
#else
PyObject *Numeric_FromLong(PyObject *obj, int precision, int scale)
{
    DBNUMERIC numeric_value;
    DBTYPEINFO info;
    DBINT conv_result;
    PyObject *strobj = PyObject_Str(obj);
    char *str;
    int num_digits;

    if (strobj == NULL)
	return NULL;
    str = PyString_AsString(strobj);
    num_digits = strlen(str);
    if (str[num_digits - 1] == 'L')
	num_digits--;
    if (precision < 0)
	precision = num_digits;
    if (precision > DB_MAX_PREC)
	precision = DB_MAX_PREC;
    info.precision = precision;
    if (scale < 0)
	scale = 0;
    info.scale = scale;
    conv_result = dbconvert_ps(NULL,
			       SQLCHAR, str, -1,
			       SQLNUMERIC, (BYTE*)&numeric_value, -1,
			       &info);
    Py_DECREF(strobj);
    if (conv_result <= 0) {
	PyErr_SetString(PyExc_TypeError, "numeric from long conversion failed");
	return NULL;
    }
    return numeric_alloc(&numeric_value);
}
#endif

#ifdef _WIN32
PyObject *Numeric_FromFloat(PyObject *obj, int precision, int scale)
{
    DBFLT8 float_value;
    DBNUMERIC numeric_value;

    float_value = PyFloat_AsDouble(obj);
    if (dbconvert(NULL,
		  SQLFLT8, (BYTE*)&float_value, -1,
		  SQLNUMERIC, (BYTE*)&numeric_value, -1) <= 0) {
	PyErr_SetString(PyExc_TypeError, "numeric from float conversion failed");
	return NULL;
    }
    return numeric_alloc(&numeric_value);
}
#else
PyObject *Numeric_FromFloat(PyObject *obj, int precision, int scale)
{
    DBFLT8 float_value;
    DBNUMERIC numeric_value;
    DBTYPEINFO info;

    if (precision < 0)
	precision = DB_MAX_PREC;
    info.precision = precision;
    if (scale < 0)
	scale = 12;
    info.scale = scale;
    float_value = PyFloat_AsDouble(obj);
    if (dbconvert_ps(NULL,
		     SQLFLT8, (BYTE*)&float_value, -1,
		     SQLNUMERIC, (BYTE*)&numeric_value, -1,
		     &info) <= 0) {
	PyErr_SetString(PyExc_TypeError, "numeric from float conversion failed");
	return NULL;
    }
    return numeric_alloc(&numeric_value);
}
#endif

#ifdef _WIN32
PyObject *Numeric_FromNumeric(PyObject *obj, int precision, int scale)
{
    DBNUMERIC numeric_value;

    if ((precision < 0 || precision == ((NumericObj*)obj)->num.precision)
	&& (scale < 0 || scale == ((NumericObj*)obj)->num.scale)) {
	Py_INCREF(obj);
	return obj;
    }
    if (dbconvert(NULL,
		  SQLNUMERIC, (BYTE*)&((NumericObj*)obj)->num, -1,
		  SQLNUMERIC, (BYTE*)&numeric_value, -1) <= 0) {
	PyErr_SetString(PyExc_TypeError, "numeric conversion failed");
	return NULL;
    }
    return numeric_alloc(&numeric_value);
}
#else
PyObject *Numeric_FromNumeric(PyObject *obj, int precision, int scale)
{
    DBNUMERIC numeric_value;
    DBTYPEINFO info;

    if ((precision < 0 || precision == ((NumericObj*)obj)->num.precision)
	&& (scale < 0 || scale == ((NumericObj*)obj)->num.scale)) {
	Py_INCREF(obj);
	return obj;
    }
    if (precision < 0)
	info.precision = ((NumericObj*)obj)->num.precision;
    else
	info.precision = precision;
    if (scale < 0)
	info.scale = ((NumericObj*)obj)->num.scale;
    else
	info.scale = scale;
    if (dbconvert_ps(NULL,
		     SQLNUMERIC, (BYTE*)&((NumericObj*)obj)->num, -1,
		     SQLNUMERIC, (BYTE*)&numeric_value, -1,
		     &info) <= 0) {
	PyErr_SetString(PyExc_TypeError, "numeric conversion failed");
	return NULL;
    }
    return numeric_alloc(&numeric_value);
}
#endif

static void Numeric_dealloc(NumericObj *self)
{
    PyMem_DEL(self);
}

static int Numeric_print(NumericObj *self, FILE *fp, int flags)
{
    char text[NUMERIC_LEN];

    numeric_as_string((PyObject*)self, text);
    fputs(text, fp);
    return 0;
}

static PyObject *Numeric_repr(NumericObj *self)
{
    char text[NUMERIC_LEN];

    numeric_as_string((PyObject*)self, text);
    return PyString_FromString(text);
}

static PyObject *Numeric_int(NumericObj *v)
{
    DBINT int_value;

    if (dbconvert(NULL,
		  SQLNUMERIC, (BYTE*)&v->num, -1,
		  SQLINT4, (BYTE*)&int_value, -1) <= 0) {
	PyErr_SetString(PyExc_TypeError, "int conversion failed");
	return NULL;
    }
    return PyInt_FromLong(int_value);
}

static PyObject *Numeric_long(NumericObj *v)
{
    char *end;
    char text[NUMERIC_LEN];

    numeric_as_string((PyObject*)v, text);
    return PyLong_FromString(text, &end, 10);
}

static PyObject *Numeric_float(NumericObj *v)
{
    DBFLT8 float_value;

    if (dbconvert(NULL,
		  SQLNUMERIC, (BYTE*)&v->num, -1,
		  SQLFLT8, (BYTE*)&float_value, -1) <= 0) {
	PyErr_SetString(PyExc_TypeError, "float conversion failed");
	return NULL;
    }
    return PyFloat_FromDouble(float_value);
}

static PyNumberMethods Numeric_as_number = {
    (binaryfunc)0,		/*nb_add*/
    (binaryfunc)0,		/*nb_subtract*/
    (binaryfunc)0,		/*nb_multiply*/
    (binaryfunc)0,		/*nb_divide*/
    (binaryfunc)0,		/*nb_remainder*/
    (binaryfunc)0,		/*nb_divmod*/
    (ternaryfunc)0,		/*nb_power*/
    (unaryfunc)0,		/*nb_negative*/
    (unaryfunc)0,		/*nb_positive*/
    (unaryfunc)0,		/*nb_absolute*/
    (inquiry)0,			/*nb_nonzero*/
    (unaryfunc)0,		/*nb_invert*/
    (binaryfunc)0,		/*nb_lshift*/
    (binaryfunc)0,		/*nb_rshift*/
    (binaryfunc)0,		/*nb_and*/
    (binaryfunc)0,		/*nb_xor*/
    (binaryfunc)0,		/*nb_or*/
    (coercion)0,		/*nb_coerce*/
    (unaryfunc)Numeric_int,	/*nb_int*/
    (unaryfunc)Numeric_long,	/*nb_long*/
    (unaryfunc)Numeric_float,	/*nb_float*/
    (unaryfunc)0,		/*nb_oct*/
    (unaryfunc)0,		/*nb_hex*/
};

#define OFFSET(x) offsetof(NumericObj, x)

static struct memberlist Numeric_memberlist[] = {
    { "precision", T_BYTE, OFFSET(num.precision), RO },
    { "scale",     T_BYTE, OFFSET(num.scale), RO },
    { NULL }			/* Sentinel */
};

static PyObject *Numeric_getattr(NumericObj *self, char *name)
{
    PyObject *rv;
	
    rv = PyMember_Get((char*)self, Numeric_memberlist, name);
    if (rv)
	return rv;
    PyErr_Clear();
    return Py_FindMethod(Numeric_methods, (PyObject *)self, name);
}

static int Numeric_setattr(NumericObj *self, char *name, PyObject *v)
{
    if (v == NULL) {
	PyErr_SetString(PyExc_AttributeError, "Cannot delete attribute");
	return -1;
    }
    return PyMember_Set((char*)self, Numeric_memberlist, name, v);
}

static char NumericType__doc__[] = 
"";

PyTypeObject NumericType = {
    PyObject_HEAD_INIT(0)
    0,				/*ob_size*/
    "NumericType",		/*tp_name*/
    sizeof(NumericObj),		/*tp_basicsize*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)Numeric_dealloc,/*tp_dealloc*/
    (printfunc)Numeric_print,	/*tp_print*/
    (getattrfunc)Numeric_getattr, /*tp_getattr*/
    (setattrfunc)Numeric_setattr, /*tp_setattr*/
    (cmpfunc)0,			/*tp_compare*/
    (reprfunc)Numeric_repr,	/*tp_repr*/
    &Numeric_as_number,		/*tp_as_number*/
    0,				/*tp_as_sequence*/
    0,				/*tp_as_mapping*/
    (hashfunc)0,		/*tp_hash*/
    (ternaryfunc)0,		/*tp_call*/
    (reprfunc)0,		/*tp_str*/

    /* Space for future expansion */
    0L,0L,0L,0L,
    NumericType__doc__		/* Documentation string */
};

char numeric_new__doc__[] =
"numeric(num, precision = -1, scale = -1) -> num\n"
"\n"
"Create a numeric object.";

/* Implement the mssqldb.numeric() method
 */
PyObject *NumericType_new(PyObject *module, PyObject *args)
{
    NumericObj *self;
    int precision, scale;
    PyObject *obj;

    self = PyObject_NEW(NumericObj, &NumericType);
    if (self == NULL)
	return NULL;

    precision = -1;
    scale = -1;
    if (PyArg_ParseTuple(args, "O|ii", &obj, &precision, &scale)) {
	PyObject *num = NULL;

	if (PyInt_Check(obj))
	    num = Numeric_FromInt(obj, precision, scale);
	else if (PyLong_Check(obj))
	    num = Numeric_FromLong(obj, precision, scale);
	else if (PyFloat_Check(obj))
	    num = Numeric_FromFloat(obj, precision, scale);
	else if (PyString_Check(obj))
	    num = Numeric_FromString(obj, precision, scale);
	else if (Numeric_Check(obj))
	    num = Numeric_FromNumeric(obj, precision, scale);
	else {
	    PyErr_SetString(PyExc_TypeError, "could not convert to Numeric");
	    return NULL;
	}
	if (num)
	    return num;
    }

    Py_DECREF(self);
    return NULL;
}

/* Used in unpickler
 */
static PyObject *numeric_constructor = NULL;

/* Register the numeric type with the copy_reg module.  This allows
 * Python to (un)pickle numeric objects.  The equivalent Python code
 * is this:
 *
 * def pickle_numeric(n):
 *     return numeric, (str(n), n.precision, n.scale)
 * 
 * copy_reg.pickle(type(numeric(1)), pickle_numeric, numeric)
 */
char pickle_numeric__doc__[] =
"pickle_numeric(n) -> numeric, (str(n), n.precision, n.scale)\n"
"\n"
"Used to pickle the numeric data type.";

/* Numeric pickling function
 */
PyObject *pickle_numeric(PyObject *module, PyObject *args)
{
    NumericObj *obj = NULL;
    PyObject *values = NULL,
	*tuple = NULL;
    char text[NUMERIC_LEN];

    if (!PyArg_ParseTuple(args, "O!", &NumericType, &obj))
	goto error;
    numeric_as_string((PyObject*)obj, text);
    if ((values = Py_BuildValue("(sii)", text,
				obj->num.precision, obj->num.scale)) == NULL)
	goto error;
    tuple = Py_BuildValue("(OO)", numeric_constructor, values);

error:
    Py_XDECREF(values);
    return tuple;
}

/* Register Numeric type pickler
 */
void copy_reg_numeric(PyObject *dict)
{
    PyObject *module = NULL,
	*pickle_func = NULL,
	*pickler = NULL,
	*obj = NULL;

    module = PyImport_ImportModule("copy_reg");
    if (module == NULL)
	goto error;
    if ((pickle_func = PyObject_GetAttrString(module, "pickle")) == NULL)
	goto error;
    if ((numeric_constructor = PyDict_GetItemString(dict, "numeric")) == NULL)
	goto error;
    if ((pickler = PyDict_GetItemString(dict, "pickle_numeric")) == NULL)
	goto error;
    obj = PyObject_CallFunction(pickle_func, "OOO",
				&NumericType, pickler, numeric_constructor);

error:
    Py_XDECREF(obj);
    Py_XDECREF(pickle_func);
    Py_XDECREF(module);
}
